/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { BinaryTreeDeletedResult, BTNCallback, BTNKey, IterationType, RBTNColor, RBTreeNodeNested, RBTreeOptions } from '../../types';
import { BST, BSTNode } from "./bst";
import { IBinaryTree } from "../../interfaces";
export declare class RBTreeNode<V = any, N extends RBTreeNode<V, N> = RBTreeNodeNested<V>> extends BSTNode<V, N> {
    color: RBTNColor;
    constructor(key: BTNKey, value?: V, color?: RBTNColor);
}
/**
 * 1. Each node is either red or black.
 * 2. The root node is always black.
 * 3. Leaf nodes are typically NIL nodes and are considered black.
 * 4. Red nodes must have black children.
 * 5. Black balance: Every path from any node to each of its leaf nodes contains the same number of black nodes.
 */
export declare class RedBlackTree<V = any, N extends RBTreeNode<V, N> = RBTreeNode<V, RBTreeNodeNested<V>>> extends BST<V, N> implements IBinaryTree<V, N> {
    constructor(options?: RBTreeOptions);
    protected _root: N;
    get root(): N;
    protected _size: number;
    get size(): number;
    NIL: N;
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    createNode(key: BTNKey, value?: V, color?: RBTNColor): N;
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C> | null | undefined, callback?: C): BinaryTreeDeletedResult<N>[];
    isNode(node: N | undefined): node is N;
    getNode<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    getNode<C extends BTNCallback<N, N>>(identifier: N | undefined, callback?: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    getNode<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    /**
     * The function returns the leftmost node in a red-black tree.
     * @param {RBTreeNode} node - The parameter "node" is of type RBTreeNode, which represents a node in
     * a Red-Black Tree.
     * @returns The leftmost node in the given RBTreeNode.
     */
    getLeftMost(node?: N): N;
    /**
     * The function returns the rightmost node in a red-black tree.
     * @param {RBTreeNode} node - The parameter "node" is of type RBTreeNode.
     * @returns the rightmost node in a red-black tree.
     */
    getRightMost(node: N): N;
    /**
     * The function returns the successor of a given node in a red-black tree.
     * @param {RBTreeNode} x - RBTreeNode - The node for which we want to find the successor.
     * @returns the successor of the given RBTreeNode.
     */
    getSuccessor(x: N): N | undefined;
    /**
     * The function returns the predecessor of a given node in a red-black tree.
     * @param {RBTreeNode} x - The parameter `x` is of type `RBTreeNode`, which represents a node in a
     * Red-Black Tree.
     * @returns the predecessor of the given RBTreeNode 'x'.
     */
    getPredecessor(x: N): N;
    clear(): void;
    protected _setRoot(v: N): void;
    /**
     * The function performs a left rotation on a red-black tree node.
     * @param {RBTreeNode} x - The parameter `x` is a RBTreeNode object.
     */
    protected _leftRotate(x: N): void;
    /**
     * The function performs a right rotation on a red-black tree node.
     * @param {RBTreeNode} x - x is a RBTreeNode, which represents the node that needs to be right
     * rotated.
     */
    protected _rightRotate(x: N): void;
    /**
     * The _fixDelete function is used to rebalance the Red-Black Tree after a node deletion.
     * @param {RBTreeNode} x - The parameter `x` is of type `RBTreeNode`, which represents a node in a
     * red-black tree.
     */
    protected _fixDelete(x: N): void;
    /**
     * The function `_rbTransplant` replaces one node in a red-black tree with another node.
     * @param {RBTreeNode} u - The parameter "u" represents a RBTreeNode object.
     * @param {RBTreeNode} v - The parameter "v" is a RBTreeNode object.
     */
    protected _rbTransplant(u: N, v: N): void;
    /**
     * The `_fixInsert` function is used to fix the red-black tree after an insertion operation.
     * @param {RBTreeNode} k - The parameter `k` is a RBTreeNode object, which represents a node in a
     * red-black tree.
     */
    protected _fixInsert(k: N): void;
}
